//-------------------------------------------------------------------------------------
// <copyright file="VirtualRelayMessageManagerSettings.h" company="LiquidRobotics">
//
// Copyright (c) 2014 Liquid Robotics, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// This software is associated with a release.  Liquid Robotics makes no
// representation that the design, interfaces or component distribution will
// remain unchanged in future releases.
//
// </copyright>
//
// <summary>
// 	Declaration of VirtualRelayMessageManagerSettings class, a serializable entity that
//  holds the various settings for the VirtualRelayMessageManager class
// </summary>
//
// <owner>Arash Ushani</owner>
//-------------------------------------------------------------------------------------

#ifndef __VirtualRelayMessageManagerSettings_H__
#define __VirtualRelayMessageManagerSettings_H__

#include "LqrSMACSharedLib.h"
#include <string>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/access.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/version.hpp>

namespace LiquidRobotics {
namespace VirtualRelay {

class SHARED_LIB_PUBLIC VirtualRelayMessageManagerSettings
{
public:

	typedef std::string string;

	VirtualRelayMessageManagerSettings();
	VirtualRelayMessageManagerSettings(const VirtualRelayMessageManagerSettings& rhs);

	const bool GetTelemetryInterrupt() const;
	void SetTelemetryInterrupt(const bool newValue);

	const bool GetAckInterrupt() const;
	void SetAckInterrupt(const bool newValue);

	const int GetTelemetryTimeoutCount() const;
	void SetTelemetryTimeoutCount(const int newValue);

	const int GetProtocolRevision() const;
	void SetProtocolRevision(const int newValue);

	const string& GetBackupFilename() const;
	void SetBackupFilename(const string& newValue);

	void Load(const string& fileName);
	void Save(const string& fileName) const;

private:

	friend class boost::serialization::access;

	template<class Archive>
	void serialize(Archive & ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(telemetryInterrupt);
		ar & BOOST_SERIALIZATION_NVP(ackInterrupt);
		ar & BOOST_SERIALIZATION_NVP(telemetryTimeoutCount);
		ar & BOOST_SERIALIZATION_NVP(protocolRevision);
		ar & BOOST_SERIALIZATION_NVP(backupFilename);
	}

	bool telemetryInterrupt;
	bool ackInterrupt;
	int telemetryTimeoutCount;
	int protocolRevision;
	string backupFilename;
};

}
}

BOOST_CLASS_VERSION(LiquidRobotics::VirtualRelay::VirtualRelayMessageManagerSettings, 1)

#endif
